use btlib::BlockMeta;
use btmsg::CallMsg;
use core::time::Duration;
use serde::{Deserialize, Serialize};
use std::fmt::Display;

pub type Inode = u64;
pub type Handle = u64;

#[derive(Serialize, Deserialize)]
pub enum FsMsg<'a> {
    #[serde(borrow)]
    Lookup(Lookup<'a>),
    #[serde(borrow)]
    Create(Create<'a>),
    Open(Open),
    Read(Read),
    #[serde(borrow)]
    Write(Write<'a>),
    Flush(Flush),
    #[serde(borrow)]
    Link(Link<'a>),
    #[serde(borrow)]
    Unlink(Unlink<'a>),
    ReadMeta(ReadMeta),
    WriteMeta(WriteMeta),
    Close(Close),
    Forget(Forget),
    Lock(Lock),
    Unlock(Unlock),
}

#[derive(Serialize, Deserialize)]
pub enum FsReply<'a> {
    Ack(()),
    Lookup(LookupReply),
    Create(CreateReply),
    Open(OpenReply),
    #[serde(borrow)]
    Read(ReadReply<'a>),
    Write(WriteReply),
    Link(LinkReply),
    ReadMeta(BlockMeta),
}

impl<'a> CallMsg<'a> for FsMsg<'a> {
    type Reply<'b> = FsReply<'b>;
}

#[derive(Debug, Serialize, Deserialize)]
pub enum FsError {
    Other,
}

impl Display for FsError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Other => write!(f, "uncategorized error"),
        }
    }
}

impl std::error::Error for FsError {}

#[derive(Serialize, Deserialize)]
pub struct Entry {
    //pub attr: stat64,
    pub attr_flags: u32,
    pub attr_timeout: Duration,
    pub entry_timeout: Duration,
}

#[derive(Serialize, Deserialize)]
pub struct Lookup<'a> {
    pub parent: Inode,
    pub name: &'a str,
}

#[derive(Serialize, Deserialize)]
pub struct LookupReply {
    pub inode: Inode,
    pub generation: u64,
    pub entry: Entry,
}

#[derive(Serialize, Deserialize)]
pub struct Create<'a> {
    pub parent: Inode,
    pub name: &'a str,
    pub flags: u32,
    pub mode: u32,
    pub umask: u32,
}

#[derive(Serialize, Deserialize)]
pub struct CreateReply {
    pub handle: Handle,
    pub entry: Entry,
}

#[derive(Serialize, Deserialize)]
pub struct Open {
    pub inode: Inode,
    pub flags: u32,
}

#[derive(Serialize, Deserialize)]
pub struct OpenReply {
    pub handle: Handle,
}

#[derive(Serialize, Deserialize)]
pub struct Read {
    pub inode: Inode,
    pub handle: Handle,
    pub offset: u64,
    pub size: u64,
}

#[derive(Serialize, Deserialize)]
pub struct ReadReply<'a> {
    pub data: &'a [u8],
}

#[derive(Serialize, Deserialize)]
pub struct Write<'a> {
    pub inode: Inode,
    pub handle: Handle,
    pub offset: u64,
    pub data: &'a [u8],
}

#[derive(Serialize, Deserialize)]
pub struct WriteReply {
    pub written: u64,
}

#[derive(Serialize, Deserialize)]
pub struct Flush {
    pub inode: Inode,
    pub handle: Handle,
}

#[derive(Serialize, Deserialize)]
pub struct Link<'a> {
    pub inode: Inode,
    pub new_parent: Inode,
    pub name: &'a str,
}

#[derive(Serialize, Deserialize)]
pub struct LinkReply {
    pub entry: Entry,
}

#[derive(Serialize, Deserialize)]
pub struct Unlink<'a> {
    pub parent: Inode,
    pub name: &'a str,
}

#[derive(Serialize, Deserialize)]
pub struct ReadMeta {
    pub inode: Inode,
    pub handle: Handle,
}

#[derive(Serialize, Deserialize)]
pub struct WriteMeta {
    pub inode: Inode,
    pub handle: Handle,
    pub meta: BlockMeta,
}

#[derive(Serialize, Deserialize)]
pub struct Close {
    pub inode: Inode,
    pub handle: Handle,
}

#[derive(Serialize, Deserialize)]
pub struct Forget {
    pub inode: Inode,
    pub count: u64,
}

#[derive(Serialize, Deserialize)]
pub struct LockDesc {
    pub offset: u64,
    pub size: u64,
    pub exclusive: bool,
}

#[derive(Serialize, Deserialize)]
pub struct Lock {
    pub inode: Inode,
    pub handle: Handle,
    pub desc: LockDesc,
}

#[derive(Serialize, Deserialize)]
pub struct Unlock {
    pub inode: Inode,
    pub handle: Handle,
}
