use crate::msg::*;

use btlib::{crypto::Creds, BlockMeta, BlockPath};
use btmsg::{receiver, MsgCallback, MsgReceived, Receiver};
use core::future::Future;
use std::{net::IpAddr, result::Result, sync::Arc};

pub trait FsProvider {
    fn lookup(&self, from: &BlockPath, msg: &Lookup) -> Result<LookupReply, FsError>;
    fn create(&self, from: &BlockPath, msg: &Create) -> Result<CreateReply, FsError>;
    fn open(&self, from: &BlockPath, msg: &Open) -> Result<OpenReply, FsError>;
    fn read(&self, from: &BlockPath, msg: &Read) -> Result<ReadReply<'_>, FsError>;
    fn write(&self, from: &BlockPath, msg: &Write) -> Result<WriteReply, FsError>;
    fn flush(&self, from: &BlockPath, msg: &Flush) -> Result<(), FsError>;
    fn link(&self, from: &BlockPath, msg: &Link) -> Result<(), FsError>;
    fn unlink(&self, from: &BlockPath, msg: &Unlink) -> Result<(), FsError>;
    fn read_meta(&self, from: &BlockPath, msg: &ReadMeta) -> Result<BlockMeta, FsError>;
    fn write_meta(&self, from: &BlockPath, msg: &WriteMeta) -> Result<(), FsError>;
    fn close(&self, from: &BlockPath, msg: &Close) -> Result<(), FsError>;
    fn forget(&self, from: &BlockPath, msg: &Forget) -> Result<(), FsError>;
    fn lock(&self, from: &BlockPath, msg: &Lock) -> Result<(), FsError>;
    fn unlock(&self, from: &BlockPath, msg: &Unlock) -> Result<(), FsError>;
}

struct ServerCallback<P> {
    provider: Arc<P>,
}

impl<P> ServerCallback<P> {
    fn new(provider: Arc<P>) -> Self {
        Self { provider }
    }
}

impl<P> Clone for ServerCallback<P> {
    fn clone(&self) -> Self {
        Self {
            provider: self.provider.clone(),
        }
    }
}

impl<P: 'static + Send + Sync + FsProvider> MsgCallback for ServerCallback<P> {
    type Arg<'de> = FsMsg<'de>;
    type CallFut<'de> = impl 'de + Future<Output = btlib::Result<()>>;
    fn call<'de>(&'de self, mut arg: MsgReceived<FsMsg<'de>>) -> Self::CallFut<'de> {
        async move {
            let provider = &self.provider;
            let reply = match arg.body() {
                FsMsg::Lookup(lookup) => FsReply::Lookup(provider.lookup(arg.from(), lookup)?),
                FsMsg::Create(create) => FsReply::Create(provider.create(arg.from(), create)?),
                FsMsg::Open(open) => FsReply::Open(provider.open(arg.from(), open)?),
                FsMsg::Read(read) => FsReply::Read(provider.read(arg.from(), read)?),
                FsMsg::Write(write) => FsReply::Write(provider.write(arg.from(), write)?),
                FsMsg::Flush(flush) => FsReply::Ack(provider.flush(arg.from(), flush)?),
                FsMsg::Link(link) => FsReply::Ack(provider.link(arg.from(), link)?),
                FsMsg::Unlink(unlink) => FsReply::Ack(provider.unlink(arg.from(), unlink)?),
                FsMsg::ReadMeta(read_meta) => {
                    FsReply::ReadMeta(provider.read_meta(arg.from(), read_meta)?)
                }
                FsMsg::WriteMeta(write_meta) => {
                    FsReply::Ack(provider.write_meta(arg.from(), write_meta)?)
                }
                FsMsg::Close(close) => FsReply::Ack(provider.close(arg.from(), close)?),
                FsMsg::Forget(forget) => FsReply::Ack(provider.forget(arg.from(), forget)?),
                FsMsg::Lock(lock) => FsReply::Ack(provider.lock(arg.from(), lock)?),
                FsMsg::Unlock(unlock) => FsReply::Ack(provider.unlock(arg.from(), unlock)?),
            };
            let mut replier = arg.take_replier().unwrap();
            replier.reply(reply).await
        }
    }
}

pub fn new_fs_server<C, P>(
    ip_addr: IpAddr,
    creds: Arc<C>,
    provider: Arc<P>,
) -> Result<impl Receiver, btlib::Error>
where
    C: 'static + Send + Sync + Creds,
    P: 'static + Send + Sync + FsProvider,
{
    receiver(ip_addr, creds, ServerCallback::new(provider))
}
